/*
 * SpanDSP - a series of DSP components for telephony
 *
 * v42bis.c
 *
 * Written by Steve Underwood <steveu@coppice.org>
 *
 * Copyright (C) 2005 Steve Underwood
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: v42bis.c,v 1.10 2005/12/25 15:08:36 steveu Exp $
 */

/* THIS IS A WORK IN PROGRESS. IT IS NOT FINISHED. */

/*! \file */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <assert.h>

#include "spandsp/telephony.h"
#include "spandsp/logging.h"
#include "spandsp/alaw_ulaw.h"
#include "spandsp/v42bis.h"

/* Fixed parameters from the spec. */
#define V42BIS_N3               8   /* Character size (bits) */
#define V42BIS_N4               256 /* Number of characters in the alphabet */
#define V42BIS_N5               (V42BIS_N4 + V42BIS_N6)  /* Index number of first dictionary entry used to store a string */
#define V42BIS_N6               3   /* Number of control codewords */

/* Control code words in compressed mode */
enum
{
    V42BIS_ETM = 0,         /* Enter transparent mode */
    V42BIS_FLUSH = 1,       /* Flush data */
    V42BIS_STEPUP = 2       /* Step up codeword size */
};

/* Command codes in transparent mode */
enum
{
    V42BIS_ECM = 0,         /* Enter compression mode */
    V42BIS_EID = 1,         /* Escape character in data */
    V42BIS_RESET = 2        /* Force reinitialisation */
    /* 7.8.2 Transition to transparent mode - TODO */
};

int v42bis_compress(v42bis_state_t *s, const uint8_t *buf, int len)
{
    int index;
    int offset;
    uint32_t octet;
    uint32_t code;
    int ptr;
    uint8_t ch;

    if ((s->v42bis_parm_p0 & 2) == 0)
    {
        /* Compression is off */
        s->frame_handler(s->frame_user_data, buf, len);
        return 0;
    }
    for (ptr = 0;  ptr < len;  ptr++)
    {
        octet = buf[ptr];
        if (s->compress.first)
        {
            s->compress.string_code = octet + V42BIS_N6;
            s->compress.first = FALSE;
        }
        else
        {
            /* Hash to find a match */
            index = (octet << (V42BIS_MAX_BITS - V42BIS_N3)) ^ s->compress.string_code;
            offset = (index == 0)  ?  1  :  V42BIS_TABLE_SIZE - index;
            for (;;)
            {
                if (s->compress.code[index] == 0xFFFF)
                    break;
                code = s->compress.code[index];
                if ((s->compress.prior_code[code] & 0x7FFF) == s->compress.string_code  &&  s->compress.node_octet[code] == octet)
                    break;
                index -= offset;
                if (index < 0)
                    index += V42BIS_TABLE_SIZE;
            }
            if (s->compress.code[index] != 0xFFFF)
            {
                /* The string was found */
                s->compress.string_code = s->compress.code[index];
                s->compress.string_length++;
            }
            else
            {
                /* The string is not in the table. */
                if (!s->compress.transparent)
                {
                    /* 7,4 Encoding - we now have the longest matchable string, and will need to output the code for it. */
                    while (s->compress.v42bis_parm_c1 >= s->compress.v42bis_parm_c3)
                    {
                        /* We need to increase the codeword size */
                        /* 7.4(a) */
                        s->compress.output_bit_buffer |= V42BIS_STEPUP << (32 - s->compress.v42bis_parm_c2 - s->compress.output_bit_count);
                        s->compress.output_bit_count += s->compress.v42bis_parm_c2;
                        while (s->compress.output_bit_count >= 8)
                        {
                            ch = s->compress.output_bit_buffer >> 24;
                            s->frame_handler(s->frame_user_data, &ch, 1);
                            s->compress.output_bit_buffer <<= 8;
                            s->compress.output_bit_count -= 8;
                        }
                        /* 7.4(b) */
                        s->compress.v42bis_parm_c2++;
                        /* 7.4(c) */
                        s->compress.v42bis_parm_c3 <<= 1;
                        /* 7.4(d) this might need to be repeated, so we loop */
                    }
                    /* 7.5 Transfer - output the last state of the string */
                    s->compress.output_bit_buffer |= s->compress.string_code << (32 - s->compress.v42bis_parm_c2 - s->compress.output_bit_count);
                    s->compress.output_bit_count += s->compress.v42bis_parm_c2;
                    while (s->compress.output_bit_count >= 8)
                    {
                        ch = s->compress.output_bit_buffer >> 24;
                        s->frame_handler(s->frame_user_data, &ch, 1);
                        s->compress.output_bit_buffer <<= 8;
                        s->compress.output_bit_count -= 8;
                    }
                }
                /* 7.6    Dictionary updating */
                /* 6.4    Add the string to the dictionary */
                /* 6.4(b) The string is not in the table. */
                if (s->compress.string_length < s->v42bis_parm_n7)
                {
                    /* 6.4(a) The length of the string is in range for adding to the dictionary */
                    s->compress.code[index] = s->compress.v42bis_parm_c1;
                    /* If the last code was a leaf, it no longer is */
                    s->compress.prior_code[s->compress.string_code] &= 0x7FFF;
                    /* The new one is definitely a leaf */
                    s->compress.prior_code[s->compress.v42bis_parm_c1] = 0x8000 | s->compress.string_code;
                    s->compress.node_octet[s->compress.v42bis_parm_c1] = octet;
                    /* 7.7    Node recovery */
                    /* 6.5    Recovering a dictionary entry to use next */
                    for (;;)
                    {
                        /* 6.5(a) and (b) */
                        if (++s->compress.v42bis_parm_c1 >= s->v42bis_parm_n2)
                            s->compress.v42bis_parm_c1 = V42BIS_N5;
                        /* 6.5(c) We need to reuse a leaf node */
                        if (!(s->compress.prior_code[s->compress.v42bis_parm_c1] & 0x8000))
                            continue;
                        if (s->compress.prior_code[s->compress.v42bis_parm_c1] == 0xFFFF)
                            break;
                        /* 6.5(d) Detach the leaf node from its parent, and re-use it */
                        index = (s->compress.node_octet[s->compress.v42bis_parm_c1] << (V42BIS_MAX_BITS - V42BIS_N3)) ^ (s->compress.prior_code[s->compress.v42bis_parm_c1] & 0x7FFF);
                        offset = (index == 0)  ?  1  :  V42BIS_TABLE_SIZE - index;
                        for (;;)
                        {
                            if (s->compress.code[index] == s->compress.v42bis_parm_c1)
                            {
                                s->compress.code[index] = 0xFFFF;
                                break;
                            }
                            index -= offset;
                            if (index < 0)
                               index += V42BIS_TABLE_SIZE;
                        }
                        s->compress.prior_code[s->compress.v42bis_parm_c1] = 0xFFFF;
                        break;
                    }
                }
                /* 7.8   Data compressibility test */
                /* 7.8.1 Transition to compressed mode - TODO */
                /* 7.8.2 Transition to transparent mode - TODO */
                /* 7.8.3 Reset function - TODO */
                s->compress.string_code = octet + V42BIS_N6;
                s->compress.string_length = 1;
            }
        }
        if (s->compress.transparent)
        {
            ch = octet;
            s->frame_handler(s->frame_user_data, &ch, 1);
        }
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/

int v42bis_compress_flush(v42bis_state_t *s)
{
    uint8_t ch;

    if (!s->compress.transparent)
    {
        /* Output the last state of the string */
        s->compress.output_bit_buffer |= s->compress.string_code << (32 - s->compress.v42bis_parm_c2 - s->compress.output_bit_count);
        s->compress.output_bit_count += s->compress.v42bis_parm_c2;
        while (s->compress.output_bit_count >= 8)
        {
            ch = s->compress.output_bit_buffer >> 24;
            s->frame_handler(s->frame_user_data, &ch, 1);
            s->compress.output_bit_buffer <<= 8;
            s->compress.output_bit_count -= 8;
        }
        /* TODO: We use a positive FLUSH at all times. It is really needed, if the
           previous step resulted in no leftover bits. */
        s->compress.output_bit_buffer |= V42BIS_FLUSH << (32 - s->compress.v42bis_parm_c2 - s->compress.output_bit_count);
        s->compress.output_bit_count += s->compress.v42bis_parm_c2;
        while (s->compress.output_bit_count > 0)
        {
            ch = s->compress.output_bit_buffer >> 24;
            s->frame_handler(s->frame_user_data, &ch, 1);
            s->compress.output_bit_buffer <<= 8;
            s->compress.output_bit_count -= 8;
        }
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/

#if 1
int v42bis_compress_dump(v42bis_state_t *s)
{
    int i;
    
    for (i = 0;  i < V42BIS_MAX_CODEWORDS;  i++)
    {
        if (s->compress.prior_code[i] != 0xFFFF)
        {
            printf("Entry %4x, prior %4x, octet %2x\n", i, s->compress.prior_code[i], s->compress.node_octet[i]);
        }
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/
#endif

int v42bis_decompress(v42bis_state_t *s, const uint8_t *buf, int len)
{
    int ptr;
    uint8_t *string;
    uint32_t code;
    uint32_t new_code;

    if ((s->v42bis_parm_p0 & 1) == 0)
    {
        /* Compression is off */
        s->data_handler(s->data_user_data, buf, len);
        return 0;
    }
    ptr = 0;
    for (;;)
    {
        if (s->decompress.transparent)
        {
            /* TODO: implement transparent mode */
            while (ptr < len)
            {
                code = buf[ptr++];
                if (code == s->decompress.escape_code)
                    s->decompress.escape_code++;
            }
            if (s->decompress.transparent)
                break;
        }
        /* Gather enough bits for at least one whole code word */
        while (s->decompress.input_bit_count <= s->decompress.v42bis_parm_c2  &&  ptr < len)
        {
            s->decompress.input_bit_count += 8;
            s->decompress.input_bit_buffer |= (uint32_t) buf[ptr++] << (32 - s->decompress.input_bit_count);
        }
        if (s->decompress.input_bit_count <= s->decompress.v42bis_parm_c2)
            break;
        new_code = s->decompress.input_bit_buffer >> (32 - s->decompress.v42bis_parm_c2);
        s->decompress.input_bit_count -= s->decompress.v42bis_parm_c2;
        s->decompress.input_bit_buffer <<= s->decompress.v42bis_parm_c2;
printf("New code %x\n", new_code);
        if (new_code < V42BIS_N6)
        {
            /* We have a control code. */
            switch (new_code)
            {
            case V42BIS_ETM:
                printf("Hit V42BIS_ETM\n");
                s->decompress.transparent = TRUE;
                break;
            case V42BIS_FLUSH:
                printf("Hit V42BIS_FLUSH\n");
                /* TODO: implement flushing */
                break;
            case V42BIS_STEPUP:
                /* We need to increase the codeword size */
                printf("Hit V42BIS_STEPUP\n");
                if (s->compress.v42bis_parm_c3 < s->v42bis_parm_n2)
                {
                    s->decompress.v42bis_parm_c2++;
                    s->decompress.v42bis_parm_c3 <<= 1;
                }
                else
                {
                    /* TODO: Report this invalid condition */
                }
                break;
            }
            continue;
        }
        if (s->decompress.first)
        {
            s->decompress.first = FALSE;
            s->decompress.octet = new_code - V42BIS_N6;
            s->decompress.decode_buf[0] = s->decompress.octet;
            s->data_handler(s->data_user_data, s->decompress.decode_buf, 1);
            s->decompress.old_code = new_code;
            continue;
        }
        /* Check the received code is valid */
        
        
        
        
        /* Start at the end of the buffer, and decode backwards */
        string = &s->decompress.decode_buf[V42BIS_MAX_STRING_SIZE - 1];
        if (new_code == s->decompress.v42bis_parm_c1)
        {
            /*
             * This code deals with special situations like "XXXXXXX". The first X goes out
             * as 0x5B. A code is then added to the encoder's dictionary for "XX". Immediately,
             * the second and third X's match this code, even though it has not yet gone into
             * the decoder's directory. We handle this by decoding the last code, and adding a
             * single character to the end of the decoded string.
             */
printf("Hit special code\n");
            *string-- = s->decompress.octet;
            code = s->decompress.old_code;
        }
        else
        {
            /* Otherwise we do a straight decode of the new code. */
            code = new_code;
        }
        /* Trace back through the octets which form the string, and output them. */
        while (code >= V42BIS_N5)
        {
            *string-- = s->decompress.node_octet[code];
            code = s->decompress.prior_code[code] & 0x7FFF;
        }
        *string = code - V42BIS_N6;
        s->decompress.octet = code - V42BIS_N6;
        /* Output the decoded string. */
printf("Output %d octets - %x %x %x %x %x %x %x %x\n", s->decompress.decode_buf + V42BIS_MAX_STRING_SIZE - string,
s->decompress.decode_buf[V42BIS_MAX_STRING_SIZE - 8],
s->decompress.decode_buf[V42BIS_MAX_STRING_SIZE - 7],
s->decompress.decode_buf[V42BIS_MAX_STRING_SIZE - 6],
s->decompress.decode_buf[V42BIS_MAX_STRING_SIZE - 5],
s->decompress.decode_buf[V42BIS_MAX_STRING_SIZE - 4],
s->decompress.decode_buf[V42BIS_MAX_STRING_SIZE - 3],
s->decompress.decode_buf[V42BIS_MAX_STRING_SIZE - 2],
s->decompress.decode_buf[V42BIS_MAX_STRING_SIZE - 1]);
        s->data_handler(s->data_user_data, string, s->decompress.decode_buf + V42BIS_MAX_STRING_SIZE - string);


        /* 6.4    Add the string to the dictionary */
        /* 6.4(b) The string is not in the table. */
        if ((s->decompress.decode_buf + V42BIS_MAX_STRING_SIZE - string) < s->v42bis_parm_n7)
        {
            /* 6.4(a) The length of the string is in range for adding to the dictionary */
printf("Adding code %x, length %d\n", new_code, (s->decompress.decode_buf + V42BIS_MAX_STRING_SIZE - string));
            /* If the last code was a leaf, it no longer is */
            s->decompress.prior_code[new_code] &= 0x7FFF;
            /* The new one is definitely a leaf */
            s->decompress.prior_code[s->decompress.v42bis_parm_c1] = 0x8000 | s->decompress.old_code;
            s->decompress.node_octet[s->decompress.v42bis_parm_c1] = s->decompress.octet;
            /* 6.5    Recovering a dictionary entry to use next */
            for (;;)
            {
                /* 6.5(a) and (b) */
                if (++s->decompress.v42bis_parm_c1 >= s->v42bis_parm_n2)
                    s->decompress.v42bis_parm_c1 = V42BIS_N5;
                /* 6.5(c) We need to reuse a leaf node */
                if (!(s->decompress.prior_code[s->decompress.v42bis_parm_c1] & 0x8000))
                    continue;
                /* 6.5(d) This is a leaf node, so re-use it */
                s->decompress.prior_code[s->decompress.v42bis_parm_c1] = 0xFFFF;
                break;
            }
        }
        s->decompress.old_code = new_code;
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/

int v42bis_decompress_flush(v42bis_state_t *s)
{
    return 0;
}
/*- End of function --------------------------------------------------------*/

#if 1
int v42bis_decompress_dump(v42bis_state_t *s)
{
    int i;
    
    for (i = 0;  i < V42BIS_MAX_CODEWORDS;  i++)
    {
        if (s->decompress.prior_code[i] != 0xFFFF)
        {
            printf("Entry %4x, prior %4x, octet %2x\n", i, s->decompress.prior_code[i], s->decompress.node_octet[i]);
        }
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/
#endif

v42bis_state_t *v42bis_init(v42bis_state_t *s,
                            int negotiated_p0,
                            int negotiated_p1,
                            int negotiated_p2,
                            v42bis_frame_handler_t frame_handler,
                            void *frame_user_data,
                            v42bis_data_handler_t data_handler,
                            void *data_user_data)
{
    int i;

    if (negotiated_p1 < 512  ||  negotiated_p1 > 65535)
        return NULL;
    if (negotiated_p2 < 6  ||  negotiated_p2 > V42BIS_MAX_STRING_SIZE)
        return NULL;
    if (s == NULL)
    {
        if ((s = (v42bis_state_t *) malloc(sizeof(*s))) == NULL)
            return NULL;
    }
    memset(s, 0, sizeof(*s));

    s->frame_handler = frame_handler;
    s->frame_user_data = frame_user_data;

    s->data_handler = data_handler;
    s->data_user_data = data_user_data;

    s->v42bis_parm_p0 = negotiated_p0;  /* default is both ways off */

    s->v42bis_parm_n1 = top_bit(negotiated_p1 - 1) + 1;
    s->v42bis_parm_n2 = negotiated_p1;
    s->v42bis_parm_n7 = negotiated_p2;

    /* 6.5 */
    s->compress.v42bis_parm_c1 =
    s->decompress.v42bis_parm_c1 = V42BIS_N5;

    s->compress.v42bis_parm_c2 =
    s->decompress.v42bis_parm_c2 = V42BIS_N3 + 1;

    s->compress.v42bis_parm_c3 =
    s->decompress.v42bis_parm_c3 = 2*V42BIS_N4;

    s->compress.first =
    s->decompress.first = TRUE;
    for (i = 0;  i < V42BIS_TABLE_SIZE;  i++)
        s->compress.code[i] = 0xFFFF;
    for (i = 0;  i < V42BIS_MAX_CODEWORDS;  i++)
    {
        s->compress.prior_code[i] =
        s->decompress.prior_code[i] = 0xFFFF;
    }
    return s;
}
/*- End of function --------------------------------------------------------*/

int v42bis_release(v42bis_state_t *s)
{
    free(s);
    return 0;
}
/*- End of function --------------------------------------------------------*/
/*- End of file ------------------------------------------------------------*/
